/* patchlevel.h */
#define VERSION 	"3.3"
#define PATCHLEVEL	7


/* sorry about this, but we should record a version somewhere,
   and I'm reluctant to muck with the items above since they're
   CMU's domain
*/
#define MODS	"Princeton mods, version 10"
