/***********************************************************
        Copyright 1996 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of CMU not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.
******************************************************************/


#ifndef DYNAMIC_H
#define DYNAMIC_H

/* ANSI function prototypes or empty arg list? */
#ifdef  __STDC__
#define P(args) args
#else
#define P(args) ()
#endif


extern time_t conftime;
extern time_t conf_lastread_time;
extern hash_tbl *cablehshtbl;
extern hash_tbl *gwhshtbl;
extern void free_cable();
extern void free_gw();
extern boolean cablelcmp();
extern boolean gwlcmp();
extern void push_ip();
extern void reuse_addr();
extern struct host *get_dynamic_ip();
extern struct host *use_addr();
extern boolean check_net P((struct in_addr *, struct in_addr *));


#endif
