/* getif.h */

/* ANSI function prototypes or empty arg list? */
#ifdef  __STDC__
#define P(args) args
#else
#define P(args) ()
#endif

extern struct in_addr *getif P((struct in_addr *));
extern void init_interfaces P((int));
extern boolean is_local P((struct in_addr *));
extern boolean is_my_ip_addr P((struct in_addr *));

/* Structure containing useful addresses associated with a single interface */
struct interface_addrs {
	struct in_addr ip_address;
	struct in_addr ip_netmask;
	struct in_addr ip_broadcast_address;
	struct in_addr ip_subnet_address;
	struct interface_addrs *next;
};

#undef P
