/* readfile.h */

#include "bootpd.h"
#include "hash.h"

#ifdef	__STDC__
#define P(args) args
#else
#define P(args) ()
#endif

extern boolean hwlookcmp P((hash_datum *, hash_datum *));
extern boolean iplookcmp P((hash_datum *, hash_datum *));
extern boolean nmcmp P((hash_datum *, hash_datum *));
extern boolean nullcmp P((hash_datum *, hash_datum *));
extern void readtab P((int));
extern void rdtab_init P((void));
extern void free_host P((hash_datum *));
extern boolean host_cmp P((struct host *, struct host *));
extern time_t modtime;
extern time_t bootptab_lastread_time;

#undef P

