#ifndef MYSYSLOG_H
#define MYSYSLOG_H

#ifdef HAVE_SYSLOG_H

#include <syslog.h>

#else

/*
 * Minimal syslog.h for systems without it  (just defines priorities).
 */

/*
 * Copyright (c) 1982, 1986 Regents of the University of California.
 * All rights reserved.  The Berkeley software License Agreement
 * specifies the terms and conditions for redistribution.
 */

/*
 *  Priorities (these are ordered)
 */
#define LOG_EMERG	0	/* system is unusable */
#define LOG_ALERT	1	/* action must be taken immediately */
#define LOG_CRIT	2	/* critical conditions */
#define LOG_ERR		3	/* error conditions */
#define LOG_WARNING	4	/* warning conditions */
#define LOG_NOTICE	5	/* normal but signification condition */
#define LOG_INFO	6	/* informational */
#define LOG_DEBUG	7	/* debug-level messages */

#endif /* HAVE_SYSLOG_H */

#endif /* MYSYSLOG_H */
